/*
 *   cdda - CD Digital Audio support
 *
 *   Copyright (C) 1993-2002  Ti Kan
 *   E-mail: xmcd@amb.org
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
#ifndef LINT
static char *_sysvipc_c_ident_ = "@(#)sysvipc.c	7.117 02/11/19";
#endif

#include "common_d/appenv.h"
#include "common_d/util.h"
#include "libdi_d/libdi.h"
#include "cdda_d/cdda.h"
#include "cdda_d/sysvipc.h"


#ifdef CDDA_SYSVIPC

#include "cdda_d/rd_scsi.h"
#include "cdda_d/rd_sol.h"
#include "cdda_d/rd_linux.h"
#include "cdda_d/rd_fbsd.h"
#include "cdda_d/rd_aix.h"
#include "cdda_d/wr_oss.h"
#include "cdda_d/wr_sol.h"
#include "cdda_d/wr_irix.h"
#include "cdda_d/wr_hpux.h"
#include "cdda_d/wr_aix.h"
#include "cdda_d/wr_alsa.h"
#include "cdda_d/wr_osf1.h"
#include "cdda_d/wr_fp.h"
#include <sys/sem.h>


extern appdata_t	app_data;
extern FILE		*errfp;
extern cdda_client_t	*cdda_clinfo;

void			*cdda_shmaddr;		/* Shared memory pointer */

STATIC int		shmid = -1,		/* Shared memory identifier */
			semid = -1,		/* Semaphores identifier */
			skip_hbchk = 0;		/* Skip heartbeat checking */
STATIC cd_state_t	*cd = NULL;		/* CD state pointer */
STATIC char		errbuf[ERR_BUF_SZ];	/* Error message buffer */


/* Call table to branch into a read-method.  The array index must
 * correspond to the cddaReadMethod parameter.
 */
struct {
	word32_t (*readinit)(void);
	bool_t	 (*read)(di_dev_t *);
	void	 (*rinfo)(char *);
} sysvipc_rd_calltbl[] = {
	{ NULL, NULL, NULL },				/* CDDA_RD_NONE */
	{ scsi_readinit, scsi_read, scsi_rinfo },	/* CDDA_RD_SCSIPT */
	{ sol_readinit, sol_read, sol_rinfo },		/* CDDA_RD_SOL */
	{ linux_readinit, linux_read, linux_rinfo },	/* CDDA_RD_LINUX */
	{ fbsd_readinit, fbsd_read, fbsd_rinfo },	/* CDDA_RD_FBSD */
	{ aix_readinit, aix_read, aix_rinfo }		/* CDDA_RD_AIX */
};


/* Call table to branch into a write-method.  The array index must
 * correspond to the cddaWriteMethod parameter.
 */
struct {
	word32_t (*writeinit)(void);
	bool_t	 (*write)(curstat_t *);
	void	 (*winfo)(char *);
} sysvipc_wr_calltbl[] = {
	{ NULL, NULL, NULL },				/* CDDA_WR_NONE */
	{ oss_writeinit, oss_write, oss_winfo },	/* CDDA_WR_OSS */
	{ sol_writeinit, sol_write, sol_winfo },	/* CDDA_WR_SOL */
	{ irix_writeinit, irix_write, irix_winfo },	/* CDDA_WR_IRIX */
	{ hpux_writeinit, hpux_write, hpux_winfo },	/* CDDA_WR_HPUX */
	{ aix_writeinit, aix_write, aix_winfo },	/* CDDA_WR_AIX */
	{ alsa_writeinit, alsa_write, alsa_winfo },	/* CDDA_WR_ALSA */
	{ osf1_writeinit, osf1_write, osf1_winfo },	/* CDDA_WR_OSF1 */
	{ fp_writeinit, fp_write, fp_winfo }		/* CDDA_WR_FP */
};


/*
 * cdda_getsem
 *	Initialize semaphores, making lock available.
 *
 * Args:
 *	None.
 *
 * Return:
 *	FALSE - failure
 *	TRUE  - success
 */
STATIC bool_t
cdda_getsem(void)
{
	semun_t	arg;

	/* Initialize semaphores */
	if ((semid = semget(SEMKEY + app_data.devnum, 3,
			    IPC_PERMS|IPC_CREAT|IPC_EXCL)) < 0) {
		/* If the semaphores exist already, use them. This
		 * should not happen if the application previously
		 * exited correctly.
		 */
		if (errno == EEXIST) {
			if ((semid = semget(SEMKEY + app_data.devnum, 3,
					    IPC_PERMS|IPC_EXCL)) < 0) {
				(void) sprintf(errbuf,
					    "cdda_getsem: semget failed "
					    "(errno=%d)",
					    errno);
				CDDA_INFO(errbuf);
				DBGPRN(DBG_GEN)(errfp, "%s\n", errbuf);
				return FALSE;
			}
		}
		else {
			(void) sprintf(errbuf,
				    "cdda_getsem: semget failed (errno=%d)",
				    errno);
			CDDA_INFO(errbuf);
			DBGPRN(DBG_GEN)(errfp, "%s\n", errbuf);
			return FALSE;
		}
	}

	/* Make lock available */
	arg.val = 1;
	if (semctl(semid, LOCK, SETVAL, arg) < 0) {
		(void) sprintf(errbuf,
			    "cdda_getsem: semctl SETVAL failed (errno=%d)",
			    errno);
		CDDA_INFO(errbuf);
		DBGPRN(DBG_GEN)(errfp, "%s\n", errbuf);
		return FALSE;
	}

	return TRUE;
}


/*
 * cdda_getshm
 *	Initialize shared memory.  We first work out how much shared
 *	memory we require and fill in a temporary cd_size_t structure
 *	along the way. We then create the shared memory segment and
 *	set the fields of *cd pointing into it.
 *
 * Args:
 *	None.
 *
 * Return:
 *	FALSE - failure
 *	TRUE  - success
 */
STATIC bool_t
cdda_getshm(void)
{
	cd_size_t	*cd_tmp_size;

	/* Store sizes here temporarily */
	cd_tmp_size = (cd_size_t *) MEM_ALLOC(
		"cd_tmp_size", sizeof(cd_size_t)
	);
	if (cd_tmp_size == NULL) {
		CDDA_FATAL(app_data.str_nomemory);
		return FALSE;
	}

	(void) memset(cd_tmp_size, 0, sizeof(cd_size_t));

	/* Sizes */
	cd_tmp_size->str_length = DEF_CDDA_STR_LEN;
	cd_tmp_size->chunk_blocks = app_data.cdda_readchkblks;

	/* Sizes depend on jitter */
	if (app_data.cdda_jitter_corr) {
		cd_tmp_size->olap_blocks = DEF_CDDA_OLAP_BLKS;
		cd_tmp_size->search_blocks = DEF_CDDA_SRCH_BLKS;
	}
	else {
		cd_tmp_size->olap_blocks = 0;
		cd_tmp_size->search_blocks = 0;
	}

	/* Chunk bytes */
	cd_tmp_size->chunk_bytes = cd_tmp_size->chunk_blocks * CDDA_BLKSZ;

	/* Buffer chunks */
	cd_tmp_size->buffer_chunks =
		(FRAME_PER_SEC * DEF_CDDA_BUFFER_SECS) /
		 cd_tmp_size->chunk_blocks;
	/* Check if we need to round up */
	if (((FRAME_PER_SEC * DEF_CDDA_BUFFER_SECS) %
	     cd_tmp_size->chunk_blocks) != 0)
		cd_tmp_size->buffer_chunks++;

	/* Buffer blocks */
	cd_tmp_size->buffer_blocks =
		cd_tmp_size->buffer_chunks * cd_tmp_size->chunk_blocks;

	/* Buffer bytes */
	cd_tmp_size->buffer_bytes =
		cd_tmp_size->buffer_chunks * cd_tmp_size->chunk_bytes;

	/* Overlap bytes */
	cd_tmp_size->olap_bytes = cd_tmp_size->olap_blocks * CDDA_BLKSZ;

	/* Search */
	cd_tmp_size->search_bytes = cd_tmp_size->search_blocks * CDDA_BLKSZ;

	/* How much shared memory do we need? */
	cd_tmp_size->size =
		sizeof(cd_size_t) + sizeof(cd_info_t) + sizeof(cd_buffer_t);

	/* Add memory for overlap */
	cd_tmp_size->size += (cd_tmp_size->search_bytes << 1);

	/* Add memory for data read */
	cd_tmp_size->size +=
		cd_tmp_size->chunk_bytes + (cd_tmp_size->olap_bytes << 1);

	/* Add memory for buffer */
	cd_tmp_size->size += cd_tmp_size->buffer_bytes;

	/* Now create the shared memory segment */
	shmid = shmget(SHMKEY + app_data.devnum, cd_tmp_size->size,
		       IPC_PERMS|IPC_CREAT|IPC_EXCL);
	if (shmid < 0) {
		/*
		 * If shared memory exists already, load it, remove it
		 * and create new one. This should not be necessary
		 * if the application previously exited correctly.
		 */
		if (errno == EEXIST) {
			shmid = shmget(SHMKEY + app_data.devnum,
				       cd_tmp_size->size, IPC_PERMS|IPC_EXCL);
			if (shmid < 0) {
				(void) sprintf(errbuf,
					    "cdda_getshm: shmget failed "
					    "(errno=%d)",
					    errno);
				CDDA_INFO(errbuf);
				DBGPRN(DBG_GEN)(errfp, "%s\n", errbuf);
				MEM_FREE(cd_tmp_size);
				return FALSE;
			}

			if (shmctl(shmid, IPC_RMID, NULL) != 0) {
				(void) sprintf(errbuf,
					    "cdda_getshm: shmctl IPC_RMID "
					    "failed (errno=%d)",
					    errno);
				CDDA_INFO(errbuf);
				DBGPRN(DBG_GEN)(errfp, "%s\n", errbuf);
				MEM_FREE(cd_tmp_size);
				return FALSE;
			}

			shmid = shmget(SHMKEY + app_data.devnum,
				       cd_tmp_size->size,
				       IPC_PERMS|IPC_CREAT|IPC_EXCL);
			if (shmid < 0) {
				(void) sprintf(errbuf,
					    "cdda_getshm: shmget failed "
					    "(errno=%d)",
					    errno);
				CDDA_INFO(errbuf);
				DBGPRN(DBG_GEN)(errfp, "%s\n", errbuf);
				MEM_FREE(cd_tmp_size);
				return FALSE;
			}
		}
		else {
			(void) sprintf(errbuf,
				    "cdda_getshm: shmget failed (errno=%d)",
				    errno);
			CDDA_INFO(errbuf);
			DBGPRN(DBG_GEN)(errfp, "%s\n", errbuf);
			MEM_FREE(cd_tmp_size);
			return FALSE;
		}
	}

	/* Attach */
	if ((cdda_shmaddr = (void *) shmat(shmid, NULL, 0)) == NULL) {
		(void) sprintf(errbuf,
				"cdda_getshm: shmat failed (errno=%d)",
				errno);
		CDDA_INFO(errbuf);
		DBGPRN(DBG_GEN)(errfp, "%s\n", errbuf);
		MEM_FREE(cd_tmp_size);
		return FALSE;
	}

	/* Now set our fields pointing into the shared memory */
	cd->cds = (cd_size_t *) cdda_shmaddr;

	/* Copy in sizes */
	(void) memcpy(cdda_shmaddr, cd_tmp_size, sizeof(cd_size_t));

	/* Info */
	cd->i = (cd_info_t *)(void *)
		((byte_t *) cdda_shmaddr + sizeof(cd_size_t));

	/* Buffer */
	cd->cdb = (cd_buffer_t *)(void *)
		((byte_t *) cd->i + sizeof(cd_info_t));

	/* Overlap */
	cd->cdb->olap = (byte_t *) ((byte_t *) cd->cdb + sizeof(cd_buffer_t));

	/* Data */
	cd->cdb->data = (byte_t *)
			((byte_t *) cd->cdb->olap +
				    (cd->cds->search_bytes << 1));

	/* Buffer */
	cd->cdb->b = (byte_t *)
		     ((byte_t *) cd->cdb->data +
				 cd->cds->chunk_bytes +
				 (cd->cds->olap_bytes << 1));

	/* Free temporary cd_size_t */
	MEM_FREE(cd_tmp_size);
	return TRUE;
}


/*
 * cdda_initshm
 *	Initialize fields of the cd_state_t structure. The cd_size_t structure
 *	has already been done in cdda_getshm(). The buffer we initialize
 *	prior to playing.
 *
 * Args:
 *	s - Pointer to the curstat_t structure.
 *
 * Return:
 *	Nothing.
 */
/*ARGSUSED*/
STATIC void
cdda_initshm(curstat_t *s)
{
	/* General state */
	cd->i->state = CDSTAT_COMPLETED;
	cd->i->chroute = app_data.ch_route;
	cd->i->vol_taper = app_data.vol_taper;
	cd->i->vol = 0;
	cd->i->vol_left = 100;
	cd->i->vol_right = 100;
	cd->i->jitter = (int) app_data.cdda_jitter_corr;
	cd->i->frm_played = 0;
	cd->i->trk_idx = 0;
	cd->i->trk_len = 0;
	cd->i->debug = app_data.debug;

	/* Start and ending LBAs for tracks */
	cd->i->start_lba = 0;
	cd->i->end_lba = 0;

	/* Initial frames per second statistic */
	cd->i->frm_per_sec = 0;

	/* Process ids */
	cd->i->reader = 0;
	cd->i->writer = 0;

	/* Reader and writer heartbeats */
	cd->i->reader_hb = 0;
	cd->i->writer_hb = 0;

	/* heartbeat timeouts */
	cd->i->reader_hb_timeout = app_data.hb_timeout;
	cd->i->writer_hb_timeout = app_data.hb_timeout * DEF_CDDA_BUFFER_SECS;
}


/*
 * cdda_sysvipc_cleanup
 *	Detaches shared memory, removes it and destroys semaphores.
 *
 * Args:
 *	None.
 *
 * Return:
 *	Nothing.
 */
STATIC void
cdda_sysvipc_cleanup(void)
{
	semun_t	arg;

	if (cd == NULL)
		return;

	DBGPRN(DBG_GEN)(errfp,
		  "\ncdda_sysvipc_cleanup: Cleaning up CDDA.\n");

	/* Detach */
	if (cd->cds != NULL && shmdt((char *) cd->cds) < 0) {
		(void) sprintf(errbuf,
			    "cdda_sysvipc_cleanup: shmdt failed (errno=%d)",
			    errno);
		CDDA_INFO(errbuf);
		DBGPRN(DBG_GEN)(errfp, "%s\n", errbuf);
	}

	/* Deallocate cd_state_t structure */
	MEM_FREE(cd);
	cd = NULL;

	/* Remove shared memory */
	if (shmid >= 0 && shmctl(shmid, IPC_RMID, NULL) < 0) {
		(void) sprintf(errbuf,
			    "cdda_sysvipc_cleanup: shmctl IPC_RMID failed "
			    "(errno=%d)",
			    errno);
		CDDA_INFO(errbuf);
		DBGPRN(DBG_GEN)(errfp, "%s\n", errbuf);
	}
	shmid = -1;

	/* Destroy semaphores */
	arg.val = 0;
	if (semid >= 0 && semctl(semid, 0, IPC_RMID, arg) < 0) {
		(void) sprintf(errbuf,
			    "cdda_sysvipc_cleanup: semctl IPC_RMID failed "
			    "(errno=%d)",
			    errno);
		CDDA_INFO(errbuf);
		DBGPRN(DBG_GEN)(errfp, "%s\n", errbuf);
	}
	semid = -1;
}


/*
 * cdda_sysvipc_init2
 *	Initialize shared memory and semaphores.
 *
 * Args:
 *	Pointer to the curstat_t structure.
 *
 * Return:
 *	TRUE - success
 *	FALSE - failure
 */
STATIC bool_t
cdda_sysvipc_init2(curstat_t *s)
{
	/* Allocate memory */
	cd = (cd_state_t *) MEM_ALLOC("cd_state_t", sizeof(cd_state_t));
	if (cd == NULL)
		/* Out of memory */
		return FALSE;

	(void) memset(cd, 0, sizeof(cd_state_t));

	/* Initialize semaphores */
	if (!cdda_getsem()) {
		cdda_sysvipc_cleanup();
		return FALSE;
	}

	/* Initialize shared memory */
	if (!cdda_getshm()) {
		cdda_sysvipc_cleanup();
		return FALSE;
	}

	/* Initialize fields */
	cdda_initshm(s);

	DBGPRN(DBG_GEN)(errfp, "\nSYSVIPC: CDDA initted.\n");
	return TRUE;
}


/*
 * cdda_sysvipc_capab
 *	Return configured CDDA capabilities
 *
 * Args:
 *	None.
 *
 * Return:
 *	Bitmask of supported CDDA read and write features
 */
word32_t
cdda_sysvipc_capab(void)
{
	word32_t	(*rinit)(void),
			(*winit)(void);

	if (app_data.cdda_rdmethod == CDDA_RD_NONE ||
	    app_data.cdda_wrmethod == CDDA_WR_NONE)
		return 0;

	if (app_data.cdda_rdmethod < CDDA_RD_NONE ||
	    app_data.cdda_rdmethod >= CDDA_RD_METHODS ||
	    app_data.cdda_wrmethod < CDDA_WR_NONE ||
	    app_data.cdda_wrmethod >= CDDA_WR_METHODS) {
		CDDA_WARNING(app_data.str_cddainit_fail);
		DBGPRN(DBG_GEN)(errfp, "Warning: %s\n",
				  app_data.str_cddainit_fail);
		return 0;
	}

	/* Call readinit and writeinit functions to determine capability */
	rinit = sysvipc_rd_calltbl[app_data.cdda_rdmethod].readinit;
	winit = sysvipc_wr_calltbl[app_data.cdda_wrmethod].writeinit;

	if (rinit == NULL || winit == NULL) {
		CDDA_WARNING(app_data.str_cddainit_fail);
		DBGPRN(DBG_GEN)(errfp, "Warning: %s\n",
				  app_data.str_cddainit_fail);
		return 0;
	}

	return ((*rinit)() | (*winit)());
}


/*
 * cdda_sysvipc_init
 *	Initialize cdda subsystem.
 *
 * Args:
 *	s - Pointer to the curstat_t structure.
 *
 * Return:
 *	FALSE - failure
 *	TRUE  - success
 */
/*ARGSUSED*/
bool_t
cdda_sysvipc_init(curstat_t *s)
{
	if (app_data.cdda_rdmethod <= CDDA_RD_NONE ||
	    app_data.cdda_rdmethod >= CDDA_RD_METHODS ||
	    app_data.cdda_wrmethod <= CDDA_WR_NONE ||
	    app_data.cdda_wrmethod >= CDDA_WR_METHODS ||
	    sysvipc_rd_calltbl[app_data.cdda_rdmethod].read == NULL ||
	    sysvipc_wr_calltbl[app_data.cdda_wrmethod].write == NULL) {
		return FALSE;
	}

	/* Defer the rest of initialization to cdda_sysvipc_init2 */
	return TRUE;
}


/*
 * cdda_sysvipc_halt
 *	Halt cdda subsystem.
 *
 * Args:
 *	devp - Read device descriptor
 *	s    - Pointer to the curstat_t structure
 *
 * Return:
 *	Nothing.
 */
void
cdda_sysvipc_halt(di_dev_t *devp, curstat_t *s)
{
	if (cd == NULL)
		return;

	/* Stop playback, if applicable */
	(void) cdda_sysvipc_stop(devp, s);

	/* Clean up IPC */
	cdda_sysvipc_cleanup();

	DBGPRN(DBG_GEN)(errfp, "\nSYSVIPC: CDDA halted.\n");
}


/*
 * cdda_sysvipc_play
 *	Start cdda play.
 *
 * Args:
 *	devp - Read device descriptor
 *	s -  Pointer to the curstat_t structure
 *	start_lba - Start logical block address
 *	end_lba   - End logical block address
 *
 * Return:
 *	TRUE - success
 *	FALSE - failure
 */
/*ARGSUSED*/
bool_t
cdda_sysvipc_play(di_dev_t *devp, curstat_t *s,
		  sword32_t start_lba, sword32_t end_lba)
{
	int	i;
	pid_t	cpid;
	semun_t	arg;
	bool_t	(*readf)(di_dev_t *),
		(*writef)(curstat_t *);

	if (cd == NULL && !cdda_sysvipc_init2(s))
		return FALSE;

	readf = sysvipc_rd_calltbl[app_data.cdda_rdmethod].read;
	writef = sysvipc_wr_calltbl[app_data.cdda_wrmethod].write;

	if (start_lba >= end_lba) {
		(void) sprintf(errbuf,
			    "cdda_sysvipc_play: "
			    "Start LBA is greater than or equal to end LBA.");
		CDDA_INFO(errbuf);
		DBGPRN(DBG_GEN)(errfp, "%s\n", errbuf);
		return FALSE;
	}

	/* If not stopped, stop first */
	if (cd->i->state != CDSTAT_COMPLETED)
		(void) cdda_sysvipc_stop(devp, s);

	/* Set status */
	cd->i->state = CDSTAT_PLAYING;

	/* Where are we starting */
	cd->i->start_lba = start_lba;
	cd->i->end_lba = end_lba;

	/* Not finished reading */
	cd->i->cdda_done = 0;

	/* Buffer pointers */
	cd->cdb->occupied = 0;
	cd->cdb->nextin = 0;
	cd->cdb->nextout = 0;

	/* Clear message buffer */
	cd->i->msgbuf[0] = '\0';

	/* Room available */
	arg.val = 1;
	if (semctl(semid, ROOM, SETVAL, arg) < 0) {
		(void) sprintf(errbuf,
			    "cdda_sysvipc_play: semctl SETVAL failed "
			    "(errno=%d)",
			    errno);
		CDDA_INFO(errbuf);
		DBGPRN(DBG_GEN)(errfp, "%s\n", errbuf);
		cd->i->state = CDSTAT_COMPLETED;
		return FALSE;
	}

	/* No data available */
	arg.val = 0;
	if (semctl(semid, DATA, SETVAL, arg) < 0) {
		(void) sprintf(errbuf,
			    "cdda_sysvipc_play: semctl SETVAL failed "
			    "(errno=%d)",
			    errno);
		CDDA_INFO(errbuf);
		DBGPRN(DBG_GEN)(errfp, "%s\n", errbuf);
		cd->i->state = CDSTAT_COMPLETED;
		return FALSE;
	}

	/* Fork CDDA reader process */
	switch (cpid = FORK()) {
	case -1:
		(void) sprintf(errbuf,
			    "cdda_sysvipc_play: fork failed (reader) "
			    "(errno=%d)",
			    errno);
		CDDA_INFO(errbuf);
		DBGPRN(DBG_GEN)(errfp, "%s\n", errbuf);
		cd->i->state = CDSTAT_COMPLETED;
		if (cd->i->writer != 0) {
			(void) kill(cd->i->writer, SIGTERM);
			cd->i->writer = 0;
		}
		return FALSE;

	case 0:
		/* Child */

		/* Close any unneeded file descriptors */
		for (i = 3; i < 10; i++) {
			if (i != devp->fd)
				(void) close(i);
		}

		/* Increase reader process scheduling priority if specified */
		if (app_data.cdda_sched & CDDA_RDPRI)
			util_highpri("Reader process");

		/* Store reader pid */
		cd->i->reader = getpid();

		/* Call read method */
		if (!(*readf)(devp))
			exit(1);

		exit(0);
		/*NOTREACHED*/

	default:
		/* Parent */
		DBGPRN(DBG_GEN)(errfp,
				"\ncdda_sysvipc_play: "
				"Started reader process pid=%d\n",
				(int) cpid);
		break;
	}

	/* Fork CDDA writer process */
	switch (cpid = FORK()) {
	case -1:
		(void) sprintf(errbuf,
			    "cdda_sysvipc_play: fork failed (writer) "
			    "(errno=%d)",
			    errno);
		CDDA_INFO(errbuf);
		DBGPRN(DBG_GEN)(errfp, "%s\n", errbuf);
		cd->i->state = CDSTAT_COMPLETED;
		if (cd->i->reader != 0) {
			(void) kill(cd->i->reader, SIGTERM);
			cd->i->reader = 0;
		}
		return FALSE;

	case 0:
		/* Child */

		/* Close any unneeded file descriptors */
		for (i = 3; i < 10; i++)
			(void) close(i);

		/* Ignore SIGUSR1 for now */
		(void) signal(SIGUSR1, SIG_IGN);

		/* Increase writer process scheduling priority if specified */
		if (app_data.cdda_sched & CDDA_WRPRI)
			util_highpri("Writer process");

#ifndef HAS_SETEUID
		/* Set to original uid/gid */
		util_set_ougid();
#endif

		/* Store writer pid */
		cd->i->writer = getpid();

		/* Call write method */
		if (!(*writef)(s))
			exit(1);

		exit(0);
		/*NOTREACHED*/

	default:
		/* Parent */
		DBGPRN(DBG_GEN)(errfp,
				"\ncdda_sysvipc_play: "
				"Started writer process pid=%d\n",
				(int) cpid);
		break;
	}

	return TRUE;
}


/*
 * cdda_sysvipc_pause_resume
 *	Pause CDDA playback.
 *
 * Args:
 *	devp   - Read device descriptor
 *	s      - Pointer to the curstat_t structure
 *	resume - Whether to resume playback
 *
 * Return:
 *	TRUE - success
 *	FALSE - failure
 */
/*ARGSUSED*/
bool_t
cdda_sysvipc_pause_resume(di_dev_t *devp, curstat_t *s, bool_t resume)
{
	if (cd == NULL && !cdda_sysvipc_init2(s))
		return FALSE;

	if (cd->i->writer != 0) {
		if (resume)
			cd->i->state = CDSTAT_PLAYING;
		else
			cd->i->state = CDSTAT_PAUSED;

		/* Signal writer if active */
		if (kill(cd->i->writer, SIGUSR1) < 0) {
			(void) sprintf(errbuf,
				    "cdda_sysvipc_pause_resume: Cannot send "
				    "SIGUSR1 to writer (errno=%d)",
				    errno);
			CDDA_INFO(errbuf);
			DBGPRN(DBG_GEN)(errfp, "%s\n", errbuf);
			return FALSE;
		}

		/* Skip heartbeat checking */
		skip_hbchk = CDDA_HB_SKIP;
	}
	return TRUE;
}


/*
 * cdda_sysvipc_stop
 *	Stop CDDA playback.
 *
 * Args:
 *	devp - Read device descriptor
 *	s    - Pointer to the curstat_t structure
 *
 * Return:
 *	TRUE - success
 *	FALSE - failure
 */
/*ARGSUSED*/
bool_t
cdda_sysvipc_stop(di_dev_t *devp, curstat_t *s)
{
	int		ret,
			saverr = 0,
			i;
	unsigned int	oa;
	pid_t		cpid;
	waitret_t	wstat;
	void		(*oh)(int);

	if (cd == NULL && !cdda_sysvipc_init2(s))
		return FALSE;

	/* Set status */
	cd->i->state = CDSTAT_COMPLETED;

	if (cd->i->writer != 0) {
		cpid = cd->i->writer;

		/* Wait for writer, blocking */
		DBGPRN(DBG_GEN)(errfp,
			"\ncdda_sysvipc_stop: Waiting for writer pid=%d\n",
			(int) cpid);

		oh = signal(SIGALRM, util_onalrm);
		oa = alarm(1);
		for (i = 0; (ret = WAITPID(cpid, &wstat, 0)) != cpid; i++) {
			(void) alarm(0);

			if (ret < 0 && errno != EINTR) {
				saverr = errno;
				break;
			}

			if (i > app_data.hb_timeout) {
				saverr = errno;
				(void) sprintf(errbuf,
					    "NOTICE: Timed out waiting for "
					    "writer process to exit (pid=%d).",
					    (int) cpid);
				CDDA_INFO(errbuf);
				DBGPRN(DBG_GEN)(errfp, "%s\n", errbuf);
				break;
			}

			(void) alarm(1);
		}
		(void) alarm(oa);
		(void) signal(SIGALRM, oh);

		if (ret < 0) {
			if (saverr != EINTR) {
				DBGPRN(DBG_GEN)(errfp,
						"waitpid on writer failed "
						"(errno=%d)\n",
						saverr);
			}
		}
		else if (WIFEXITED(wstat)) {
			ret = WEXITSTATUS(wstat);
			DBGPRN(DBG_GEN)(errfp, "Writer exited, status %d\n",
					ret);

			/* Display message, if any */
			if (ret != 0 && cd->i->msgbuf[0] != '\0')
				CDDA_INFO(cd->i->msgbuf);
		}
		else if (WIFSIGNALED(wstat)) {
			DBGPRN(DBG_GEN)(errfp, "Writer killed, signal %d\n",
					WTERMSIG(wstat));
		}

		cd->i->writer = 0;
	}

	if (cd->i->reader != 0) {
		cpid = cd->i->reader;

		/* Wait for reader, blocking */
		DBGPRN(DBG_GEN)(errfp,
			"\ncdda_sysvipc_stop: Waiting for reader pid=%d\n",
			(int) cpid);

		oh = signal(SIGALRM, util_onalrm);
		oa = alarm(1);
		for (i = 0; (ret = WAITPID(cpid, &wstat, 0)) != cpid; i++) {
			(void) alarm(0);

			if (ret < 0 && errno != EINTR) {
				saverr = errno;
				break;
			}

			if (i > app_data.hb_timeout) {
				saverr = errno;
				(void) sprintf(errbuf,
					    "NOTICE: Timed out waiting for "
					    "reader process to exit (pid=%d).",
					    (int) cpid);
				CDDA_INFO(errbuf);
				DBGPRN(DBG_GEN)(errfp, "%s\n", errbuf);
				break;
			}

			(void) alarm(1);
		}
		(void) alarm(oa);
		(void) signal(SIGALRM, oh);

		if (ret < 0) {
			if (saverr != EINTR) {
				DBGPRN(DBG_GEN)(errfp,
						"waitpid on reader failed "
						"(errno=%d)\n",
						saverr);
			}
		}
		else if (WIFEXITED(wstat)) {
			ret = WEXITSTATUS(wstat);
			DBGPRN(DBG_GEN)(errfp, "Reader exited, status %d\n",
				        ret);

			/* Display message, if any */
			if (ret != 0 && cd->i->msgbuf[0] != '\0')
				CDDA_INFO(cd->i->msgbuf);
		}
		else if (WIFSIGNALED(wstat)) {
			DBGPRN(DBG_GEN)(errfp, "Reader killed, signal %d\n",
				        WTERMSIG(wstat));
		}

		cd->i->reader = 0;
	}

	/* Reset states */
	cdda_initshm(s);

	return TRUE;
}


/*
 * cdda_sysvipc_vol
 *	Change volume setting.
 *
 * Args:
 *	devp  - Read device descriptor
 *	s     - Pointer to the curstat_t structure
 *	vol   - Desired volume level
 *	query - Whether querying or setting the volume
 *
 * Return:
 *	The volume setting, or -1 on failure.
 */
/*ARGSUSED*/
int
cdda_sysvipc_vol(di_dev_t *devp, curstat_t *s, int vol, bool_t query)
{
	if (cd == NULL && !cdda_sysvipc_init2(s))
		return -1;

	if (query) {
		s->level_left = (byte_t) cd->i->vol_left;
		s->level_right = (byte_t) cd->i->vol_right;
		return (cd->i->vol);
	}

	cd->i->vol_taper = app_data.vol_taper;
	cd->i->vol = vol;
	cd->i->vol_left = (int) s->level_left;
	cd->i->vol_right = (int) s->level_right;

	/* Signal writer if active */
	if (cd->i->writer != 0) {
		if (kill(cd->i->writer, SIGUSR1) < 0) {
			(void) sprintf(errbuf,
				    "cdda_sysvipc_vol: Cannot send "
				    "SIGUSR1 to writer (errno=%d)",
				    errno);
			CDDA_INFO(errbuf);
			DBGPRN(DBG_GEN)(errfp, "%s\n", errbuf);
			return -1;
		}
	}

	return (vol);
}


/*
 * cdda_sysvipc_chroute
 *	Change channel routing setting.
 *
 * Args:
 *	devp - Read device descriptor
 *	s    - Pointer to the curstat_t structure
 *
 * Return:
 *	TRUE - success
 *	FALSE - failure
 */
/*ARGSUSED*/
bool_t
cdda_sysvipc_chroute(di_dev_t *devp, curstat_t *s)
{
	if (cd != NULL) {
		cd->i->chroute = app_data.ch_route;
		return TRUE;
	}
	else
		return FALSE;
}


/*
 * cdda_sysvipc_getstatus
 *	Get CDDA playback status.
 *
 * Args:
 *	devp - Read device descriptor
 *	s    - Pointer to curstat_t structure
 *	sp - CDDA status return structure
 *
 * Return:
 *	TRUE  - success
 *	FALSE - failure
 */
/*ARGSUSED*/
bool_t
cdda_sysvipc_getstatus(di_dev_t *devp, curstat_t *s, cdstat_t *sp)
{
	int		i;
	time_t		now;
	bool_t		reader_fail,
			writer_fail;
	static int	hbint = 0,
			hbcnt = 0;

	if (cd == NULL && !cdda_sysvipc_init2(s))
		return FALSE;

	/* Current playback status */
	sp->status = cd->i->state;
	if (sp->status == CDSTAT_COMPLETED)
		(void) cdda_sysvipc_stop(devp, s);

	/* Current playback location */
	sp->abs_addr.addr = cd->i->start_lba + cd->i->frm_played;

	util_blktomsf(
		sp->abs_addr.addr,
		&sp->abs_addr.min,
		&sp->abs_addr.sec,
		&sp->abs_addr.frame,
		MSF_OFFSET
	);

	i = cd->i->trk_idx;
	if (sp->abs_addr.addr >= s->trkinfo[i].addr &&
	    sp->abs_addr.addr < s->trkinfo[i+1].addr) {
			sp->track = s->trkinfo[i].trkno;

			sp->rel_addr.addr =
				sp->abs_addr.addr - s->trkinfo[i].addr;

			util_blktomsf(
				sp->rel_addr.addr,
				&sp->rel_addr.min,
				&sp->rel_addr.sec,
				&sp->rel_addr.frame,
				0
			);
	}
	else for (i = 0; i < (int) s->tot_trks; i++) {
		if (sp->abs_addr.addr >= s->trkinfo[i].addr &&
		    sp->abs_addr.addr < s->trkinfo[i+1].addr) {
			sp->track = s->trkinfo[i].trkno;

			sp->rel_addr.addr =
				sp->abs_addr.addr - s->trkinfo[i].addr;

			util_blktomsf(
				sp->rel_addr.addr,
				&sp->rel_addr.min,
				&sp->rel_addr.sec,
				&sp->rel_addr.frame,
				0
			);
			break;
		}
	}

	sp->index = 1;	/* Index number not supported in this mode */

	/* Current volume and balance */
	sp->level = (byte_t) cd->i->vol;
	sp->level_left = (byte_t) cd->i->vol_left;
	sp->level_right = (byte_t) cd->i->vol_right;

	sp->tot_frm = cd->i->end_lba - cd->i->start_lba + 1;
	sp->frm_played = cd->i->frm_played;
	sp->frm_per_sec = cd->i->frm_per_sec;

	/* Initialize heartbeat checking interval */
	if (hbint == 0)
		hbcnt = hbint = (1000 / app_data.stat_interval) + 1;

	/* Check reader and writer heartbeats */
	if (sp->status == CDSTAT_PLAYING && --hbcnt == 0 &&
	    cd->i->reader != 0 && cd->i->writer != 0 &&
	    cd->i->reader_hb != 0 && cd->i->writer_hb != 0) {
		hbcnt = hbint;
		now = time(NULL);

		if (skip_hbchk > 0) {
		    /* Skip heartbeat checking for a few iterations after
		     * resuming from pause, to let the reader and writer
		     * threads to catch up.
		     */
		    reader_fail = writer_fail = FALSE;
		    skip_hbchk--;
		}
		else {
		    reader_fail = (bool_t)
			((now - cd->i->reader_hb) > cd->i->reader_hb_timeout);
		    writer_fail = (bool_t)
			((now - cd->i->writer_hb) > cd->i->writer_hb_timeout);
		}

		if (reader_fail || writer_fail) {
			/* Reader or writer is hung or died */
			(void) kill(cd->i->reader, SIGTERM);
			(void) kill(cd->i->writer, SIGTERM);

			(void) cdda_sysvipc_stop(devp, s);

			(void) sprintf(errbuf,
				       "CDDA %s thread failure!",
				       writer_fail ? "writer" : "reader");
			CDDA_INFO(errbuf);
			DBGPRN(DBG_GEN)(errfp, "%s\n", errbuf);
		}
	}

	return TRUE;
}


/*
 * cdda_sysvipc_debug
 *	Debug level change notification function
 *
 * Args:
 *	lev - New debug level
 *
 * Return:
 *	Nothing.
 */
void
cdda_sysvipc_debug(word32_t lev)
{
	if (cd != NULL)
		cd->i->debug = lev;
}



/*
 * cdda_sysvipc_info
 *	Append CDDA read and write method information to supplied string.
 *
 * Args:
 *	str - The string to append to.
 *
 * Return:
 *	Nothing.
 */
void
cdda_sysvipc_info(char *str)
{
	void	(*rinfo)(char *),
		(*winfo)(char *);

	(void) strcat(str, "SYSV IPC\n");

	(void) strcat(str, "    Extract:  ");
	if (app_data.cdda_rdmethod <= CDDA_RD_NONE ||
	    app_data.cdda_rdmethod >= CDDA_RD_METHODS) {
		(void) strcat(str, "not configured\n");	
	}
	else {
		rinfo = sysvipc_rd_calltbl[app_data.cdda_rdmethod].rinfo;
		if (rinfo == NULL)
			(void) strcat(str, "not configured\n");	
		else
			(*rinfo)(str);
	}

	(void) strcat(str, "    Playback: ");
	if (app_data.cdda_wrmethod <= CDDA_WR_NONE ||
	    app_data.cdda_wrmethod >= CDDA_WR_METHODS) {
		(void) strcat(str, "not configured\n");	
	}
	else {
		winfo = sysvipc_wr_calltbl[app_data.cdda_wrmethod].winfo;
		if (winfo == NULL)
			(void) strcat(str, "not configured\n");	
		else
			(*winfo)(str);
	}
}


/*
 * cdda_rw_initipc
 *	Retrieves shared memory and semaphores. Sets up our pointers
 *	from cd into shared memory.  Used by the reader/writer child
 *	processes.
 *
 * Args:
 *	cd - Pointer to the cd_state_t structure to be filled in
 *
 * Return:
 *	The IPC semaphore ID, or -1 if failed.
 */
int
cdda_rw_initipc(cd_state_t *cdp)
{
	int	id;

	if (cdda_shmaddr == NULL) {
		DBGPRN(DBG_GEN)(errfp,
			"gen_write_initipc: No shared memory!\n");
		return -1;
	}

	/* Now set our fields pointing into the shared memory */
	cdp->cds = (cd_size_t *) cdda_shmaddr;

	/* Info */
	cdp->i = (cd_info_t *)(void *)
		((byte_t *) cdda_shmaddr + sizeof(cd_size_t));

	/* Buffer */
	cdp->cdb = (cd_buffer_t *)(void *)
		((byte_t *) cd->i + sizeof(cd_info_t));

	/* Overlap */
	cdp->cdb->olap = (byte_t *) ((byte_t *) cd->cdb + sizeof(cd_buffer_t));

	/* Data */
	cdp->cdb->data = (byte_t *)
			 ((byte_t *) cd->cdb->olap +
				    (cd->cds->search_bytes << 1));

	/* Buffer */
	cdp->cdb->b = (byte_t *)
		      ((byte_t *) cd->cdb->data + cd->cds->chunk_bytes +
				 (cd->cds->olap_bytes << 1));

	/* Semaphores */
	if ((id = semget(SEMKEY + app_data.devnum, 3,
			 IPC_PERMS|IPC_EXCL)) < 0) {
		DBGPRN(DBG_GEN)(errfp,
			"gen_write_initipc: semget failed: errno=%d\n",
			errno);
	}

	return (id);
}


/*
 * cdda_rd_corrjitter
 *	Correct for jitter. The olap array contains cdp->cds->search_bytes * 2
 *	of data and the last data written to the device is situated at
 *	olap[cdp->cds->search_bytes]. We try to match the last data read,
 *	which starts at data[cdp->cds->olap_bytes], around the center of olap.
 *	Once matched we may have to copy some data from olap into data
 *	(overshoot) or skip some of the audio in data (undershoot). The
 *	function returns a pointer into data indicating from where we should
 *	start writing.
 *	Used by the reader child process only.
 *
 * Args:
 *	None.
 *
 * Return:
 *	Pointer into data
 */
byte_t *
cdda_rd_corrjitter(cd_state_t *cdp)
{
	int		idx,
			d;
	byte_t		*i,
			*j,
			*l;

	/* If jitter correction off, return */
	if (!cdp->i->jitter)
		return (cdp->cdb->data);

	/* Find a match in olap, if we can */
	i = cdp->cdb->olap;
	l = &cdp->cdb->olap[(cdp->cds->search_bytes << 1) - 1];
	j = &cdp->cdb->data[cdp->cds->olap_bytes];

	do {
		i = memchr(i, (int) *j, l - i);
		if (i != NULL) {
			if ((i - cdp->cdb->olap) % 2 == 0) {
				d = memcmp(j, i, cdp->cds->str_length);
				if (d == 0)
					break;
			}
			i++;
		}
	} while (i != NULL && i < l);

	/* Did we find a match? */
	if (i == NULL || i >= l) {
		DBGPRN(DBG_GEN)(errfp,
		    "cdda_rd_corrjitter: Could not correct for jitter.\n");
		idx = 0;
	}
	else {
		idx = i - cdp->cdb->olap - cdp->cds->search_bytes;
	}

	/* Update pointer */
	j -= idx;

	/* Match on RHS, copy some olap into data */
	if (idx > 0)
		(void) memcpy(j, cdp->cdb->olap + cdp->cds->search_bytes, idx);

	return (j);
}


/*
 * cdda_wr_setcurtrk
 *	Set the track index and track length based on the current playing
 *	location.
 *
 * Args:
 *	cdp - Pointer to the cd_state_t structure
 *
 * Return:
 *	Nothing.
 */
void
cdda_wr_setcurtrk(cd_state_t *cdp)
{
	int		i;
	sword32_t	lba;
	curstat_t	*s = cdda_clinfo->curstat_addr();

	if (cdp->i->state != CDSTAT_PLAYING && cdp->i->state != CDSTAT_PAUSED)
		return;

	lba = cdp->i->start_lba + cdp->i->frm_played;

	i = cdp->i->trk_idx;
	if (i >= 0 &&
	    lba >= s->trkinfo[i].addr && lba < s->trkinfo[i+1].addr) {
		cdp->i->trk_idx = i;
		cdp->i->trk_len = CDDA_BLKSZ *
			(s->trkinfo[i+1].addr - s->trkinfo[i].addr + 1);
		return;
	}

	for (i = 0; i < (int) s->tot_trks; i++) {
		if (lba >= s->trkinfo[i].addr &&
		    lba < s->trkinfo[i+1].addr) {
			cdp->i->trk_idx = i;
			cdp->i->trk_len = CDDA_BLKSZ *
			    (s->trkinfo[i+1].addr - s->trkinfo[i].addr + 1);
		}
	}
}


#endif	/* CDDA_SYSVIPC */

